@@ -13,6 +13,7 @@ object and then create the configurable objects, passing the config to them.
 # Distributed under the terms of the Modified BSD License.
 
 import atexit
+from copy import deepcopy
 import glob
 import logging
 import os
@@ -25,8 +26,10 @@ from IPython.core import release, crashhandler
 from IPython.core.profiledir import ProfileDir, ProfileDirError
 from IPython.paths import get_ipython_dir, get_ipython_package_dir
 from IPython.utils.path import ensure_dir_exists
-from IPython.utils import py3compat
-from traitlets import List, Unicode, Type, Bool, Dict, Set, Instance, Undefined
+from traitlets import (
+    List, Unicode, Type, Bool, Set, Instance, Undefined,
+    default, observe,
+)
 
 if os.name == 'nt':
     programdata = os.environ.get('PROGRAMDATA', None)
@@ -41,6 +44,24 @@ else:
     ]
 
 
+ENV_CONFIG_DIRS = []
+_env_config_dir = os.path.join(sys.prefix, 'etc', 'ipython')
+if _env_config_dir not in SYSTEM_CONFIG_DIRS:
+    # only add ENV_CONFIG if sys.prefix is not already included
+    ENV_CONFIG_DIRS.append(_env_config_dir)
+
+
+_envvar = os.environ.get('IPYTHON_SUPPRESS_CONFIG_ERRORS')
+if _envvar in {None, ''}:
+    IPYTHON_SUPPRESS_CONFIG_ERRORS = None
+else:
+    if _envvar.lower() in {'1','true'}:
+        IPYTHON_SUPPRESS_CONFIG_ERRORS = True
+    elif _envvar.lower() in {'0','false'} :
+        IPYTHON_SUPPRESS_CONFIG_ERRORS = False
+    else:
+        sys.exit("Unsupported value for environment variable: 'IPYTHON_SUPPRESS_CONFIG_ERRORS' is set to '%s' which is none of  {'0', '1', 'false', 'true', ''}."% _envvar )
+
 # aliases and flags
 
 base_aliases = {
@@ -80,12 +101,12 @@ class ProfileAwareConfigLoader(PyFileConfigLoader):
 
 class BaseIPythonApplication(Application):
 
-    name = Unicode(u'ipython')
+    name = u'ipython'
     description = Unicode(u'IPython: an enhanced interactive Python shell.')
     version = Unicode(release.version)
 
-    aliases = Dict(base_aliases)
-    flags = Dict(base_flags)
+    aliases = base_aliases
+    flags = base_flags
     classes = List([ProfileDir])
     
     # enable `load_subconfig('cfg.py', profile='name')`
@@ -96,11 +117,13 @@ class BaseIPythonApplication(Application):
     config_file_specified = Set()
 
     config_file_name = Unicode()
+    @default('config_file_name')
     def _config_file_name_default(self):
         return self.name.replace('-','_') + u'_config.py'
-    def _config_file_name_changed(self, name, old, new):
-        if new != old:
-            self.config_file_specified.add(new)
+    @observe('config_file_name')
+    def _config_file_name_changed(self, change):
+        if change['new'] != change['old']:
+            self.config_file_specified.add(change['new'])
 
     # The directory that contains IPython's builtin profiles.
     builtin_profile_dir = Unicode(
@@ -108,15 +131,19 @@ class BaseIPythonApplication(Application):
     )
     
     config_file_paths = List(Unicode())
+    @default('config_file_paths')
     def _config_file_paths_default(self):
-        return [py3compat.getcwd()]
+        return []
 
-    extra_config_file = Unicode(config=True,
+    extra_config_file = Unicode(
     help="""Path to an extra config file to load.
     
     If specified, load this config file in addition to any other IPython config.
-    """)
-    def _extra_config_file_changed(self, name, old, new):
+    """).tag(config=True)
+    @observe('extra_config_file')
+    def _extra_config_file_changed(self, change):
+        old = change['old']
+        new = change['new']
         try:
             self.config_files.remove(old)
         except ValueError:
@@ -124,30 +151,37 @@ class BaseIPythonApplication(Application):
         self.config_file_specified.add(new)
         self.config_files.append(new)
 
-    profile = Unicode(u'default', config=True,
+    profile = Unicode(u'default',
         help="""The IPython profile to use."""
-    )
-
-    def _profile_changed(self, name, old, new):
+    ).tag(config=True)
+    
+    @observe('profile')
+    def _profile_changed(self, change):
         self.builtin_profile_dir = os.path.join(
-                get_ipython_package_dir(), u'config', u'profile', new
+                get_ipython_package_dir(), u'config', u'profile', change['new']
         )
 
-    ipython_dir = Unicode(config=True,
+    ipython_dir = Unicode(
         help="""
         The name of the IPython directory. This directory is used for logging
         configuration (through profiles), history storage, etc. The default
         is usually $HOME/.ipython. This option can also be specified through
         the environment variable IPYTHONDIR.
         """
-    )
+    ).tag(config=True)
+    @default('ipython_dir')
     def _ipython_dir_default(self):
         d = get_ipython_dir()
-        self._ipython_dir_changed('ipython_dir', d, d)
+        self._ipython_dir_changed({
+            'name': 'ipython_dir',
+            'old': d,
+            'new': d,
+        })
         return d
     
     _in_init_profile_dir = False
     profile_dir = Instance(ProfileDir, allow_none=True)
+    @default('profile_dir')
     def _profile_dir_default(self):
         # avoid recursion
         if self._in_init_profile_dir:
@@ -156,26 +190,29 @@ class BaseIPythonApplication(Application):
         self.init_profile_dir()
         return self.profile_dir
 
-    overwrite = Bool(False, config=True,
-        help="""Whether to overwrite existing config files when copying""")
-    auto_create = Bool(False, config=True,
-        help="""Whether to create profile dir if it doesn't exist""")
+    overwrite = Bool(False,
+        help="""Whether to overwrite existing config files when copying"""
+    ).tag(config=True)
+    auto_create = Bool(False,
+        help="""Whether to create profile dir if it doesn't exist"""
+    ).tag(config=True)
 
     config_files = List(Unicode())
+    @default('config_files')
     def _config_files_default(self):
         return [self.config_file_name]
 
-    copy_config_files = Bool(False, config=True,
+    copy_config_files = Bool(False,
         help="""Whether to install the default config files into the profile dir.
         If a new profile is being created, and IPython contains config files for that
         profile, then they will be staged into the new directory.  Otherwise,
         default config files will be automatically generated.
-        """)
+        """).tag(config=True)
     
-    verbose_crash = Bool(False, config=True,
+    verbose_crash = Bool(False,
         help="""Create a massive crash report when IPython encounters what may be an
         internal error.  The default is to append a short message to the
-        usual traceback""")
+        usual traceback""").tag(config=True)
 
     # The class to use as the crash handler.
     crash_handler_class = Type(crashhandler.CrashHandler)
@@ -185,7 +222,7 @@ class BaseIPythonApplication(Application):
         super(BaseIPythonApplication, self).__init__(**kwargs)
         # ensure current working directory exists
         try:
-            directory = py3compat.getcwd()
+            os.getcwd()
         except:
             # exit if cwd doesn't exist
             self.log.error("Current working directory doesn't exist.")
@@ -194,6 +231,16 @@ class BaseIPythonApplication(Application):
     #-------------------------------------------------------------------------
     # Various stages of Application creation
     #-------------------------------------------------------------------------
+    
+    deprecated_subcommands = {}
+    
+    def initialize_subcommand(self, subc, argv=None):
+        if subc in self.deprecated_subcommands:
+            self.log.warning("Subcommand `ipython {sub}` is deprecated and will be removed "
+                             "in future versions.".format(sub=subc))
+            self.log.warning("You likely want to use `jupyter {sub}` in the "
+                             "future".format(sub=subc))
+        return super(BaseIPythonApplication, self).initialize_subcommand(subc, argv)
 
     def init_crash_handler(self):
         """Create a crash handler, typically setting sys.excepthook to it."""
@@ -214,17 +261,16 @@ class BaseIPythonApplication(Application):
             return self.crash_handler(etype, evalue, tb)
         else:
             return crashhandler.crash_handler_lite(etype, evalue, tb)
-    
-    def _ipython_dir_changed(self, name, old, new):
+
+    @observe('ipython_dir')
+    def _ipython_dir_changed(self, change):
+        old = change['old']
+        new = change['new']
         if old is not Undefined:
-            str_old = py3compat.cast_bytes_py2(os.path.abspath(old),
-                sys.getfilesystemencoding()
-            )
+            str_old = os.path.abspath(old)
             if str_old in sys.path:
                 sys.path.remove(str_old)
-        str_path = py3compat.cast_bytes_py2(os.path.abspath(new),
-            sys.getfilesystemencoding()
-        )
+        str_path = os.path.abspath(new)
         sys.path.append(str_path)
         ensure_dir_exists(new)
         readme = os.path.join(new, 'README')
@@ -240,18 +286,33 @@ class BaseIPythonApplication(Application):
                 self.log.error("couldn't create path %s: %s", path, e)
         self.log.debug("IPYTHONDIR set to: %s" % new)
 
-    def load_config_file(self, suppress_errors=True):
+    def load_config_file(self, suppress_errors=IPYTHON_SUPPRESS_CONFIG_ERRORS):
         """Load the config file.
 
         By default, errors in loading config are handled, and a warning
         printed on screen. For testing, the suppress_errors option is set
         to False, so errors will make tests fail.
+
+        `suppress_errors` default value is to be `None` in which case the
+        behavior default to the one of `traitlets.Application`.
+
+        The default value can be set :
+           - to `False` by setting 'IPYTHON_SUPPRESS_CONFIG_ERRORS' environment variable to '0', or 'false' (case insensitive).
+           - to `True` by setting 'IPYTHON_SUPPRESS_CONFIG_ERRORS' environment variable to '1' or 'true' (case insensitive).
+           - to `None` by setting 'IPYTHON_SUPPRESS_CONFIG_ERRORS' environment variable to '' (empty string) or leaving it unset.
+
+        Any other value are invalid, and will make IPython exit with a non-zero return code.
         """
+
+
         self.log.debug("Searching path %s for config files", self.config_file_paths)
         base_config = 'ipython_config.py'
         self.log.debug("Attempting to load config file: %s" %
                        base_config)
         try:
+            if suppress_errors is not None:
+                old_value = Application.raise_config_file_errors
+                Application.raise_config_file_errors = not suppress_errors;
             Application.load_config_file(
                 self,
                 base_config,
@@ -261,6 +322,8 @@ class BaseIPythonApplication(Application):
             # ignore errors loading parent
             self.log.debug("Config file %s not found", base_config)
             pass
+        if suppress_errors is not None:
+            Application.raise_config_file_errors = old_value
         
         for config_file_name in self.config_files:
             if not config_file_name or config_file_name == base_config:
@@ -276,7 +339,7 @@ class BaseIPythonApplication(Application):
             except ConfigFileNotFound:
                 # Only warn if the default config file was NOT being used.
                 if config_file_name in self.config_file_specified:
-                    msg = self.log.warn
+                    msg = self.log.warning
                 else:
                     msg = self.log.debug
                 msg("Config file not found, skipping: %s", config_file_name)
@@ -284,7 +347,7 @@ class BaseIPythonApplication(Application):
                 # For testing purposes.
                 if not suppress_errors:
                     raise
-                self.log.warn("Error loading config file: %s" %
+                self.log.warning("Error loading config file: %s" %
                               self.config_file_name, exc_info=True)
 
     def init_profile_dir(self):
@@ -343,6 +406,7 @@ class BaseIPythonApplication(Application):
 
     def init_config_files(self):
         """[optionally] copy default config files into profile dir."""
+        self.config_file_paths.extend(ENV_CONFIG_DIRS)
         self.config_file_paths.extend(SYSTEM_CONFIG_DIRS)
         # copy config files
         path = self.builtin_profile_dir
@@ -351,7 +415,7 @@ class BaseIPythonApplication(Application):
 
             cfg = self.config_file_name
             if path and os.path.exists(os.path.join(path, cfg)):
-                self.log.warn("Staging %r from %s into %r [overwrite=%s]"%(
+                self.log.warning("Staging %r from %s into %r [overwrite=%s]"%(
                         cfg, src, self.profile_dir.location, self.overwrite)
                 )
                 self.profile_dir.copy_config_file(cfg, path=path, overwrite=self.overwrite)
@@ -366,7 +430,7 @@ class BaseIPythonApplication(Application):
                 cfg = os.path.basename(fullpath)
                 if self.profile_dir.copy_config_file(cfg, path=path, overwrite=False):
                     # file was copied
-                    self.log.warn("Staging bundled %s from %s into %r"%(
+                    self.log.warning("Staging bundled %s from %s into %r"%(
                             cfg, self.profile, self.profile_dir.location)
                     )
 
@@ -376,7 +440,7 @@ class BaseIPythonApplication(Application):
         s = self.generate_config_file()
         fname = os.path.join(self.profile_dir.location, self.config_file_name)
         if self.overwrite or not os.path.exists(fname):
-            self.log.warn("Generating default config file: %r"%(fname))
+            self.log.warning("Generating default config file: %r"%(fname))
             with open(fname, 'w') as f:
                 f.write(s)
 
@@ -388,10 +452,11 @@ class BaseIPythonApplication(Application):
         if self.subapp is not None:
             # stop here if subapp is taking over
             return
-        cl_config = self.config
+        # save a copy of CLI config to re-load after config files
+        # so that it has highest priority
+        cl_config = deepcopy(self.config)
         self.init_profile_dir()
         self.init_config_files()
         self.load_config_file()
         # enforce cl-opts override configfile opts:
         self.update_config(cl_config)
-
